""" Generator functions to calculate errors in the driver """

from dune.perftool.generation import (cached,
                                      include_file,
                                      preamble,
                                      )
from dune.perftool.options import get_option
from dune.perftool.pdelab.driver import (get_formdata,
                                         get_trial_element,
                                         preprocess_leaf_data,
                                         )
from dune.perftool.pdelab.driver.gridfunctionspace import name_gfs
from dune.perftool.pdelab.driver.interpolate import (interpolate_vector,
                                                     name_boundary_function,
                                                     )
from dune.perftool.pdelab.driver.solve import (define_vector,
                                               name_vector,
                                               )


@preamble
def define_test_fail_variable(name):
    return 'bool {}(false);'.format(name)


def name_test_fail_variable():
    name = "testfail"
    define_test_fail_variable(name)
    return name


def name_exact_solution():
    name = "solution"
    define_vector(name, get_formdata())
    interpolate_exact_solution(name)
    return name


def interpolate_exact_solution(name):
    element = get_trial_element()
    is_dirichlet = preprocess_leaf_data(element, "is_dirichlet")
    gfs = name_gfs(element, is_dirichlet)
    sol = preprocess_leaf_data(element, "exact_solution")
    func = name_boundary_function(element, sol)
    interpolate_vector(func, gfs, name)


@preamble
def compare_dofs():
    v = name_vector(get_formdata())
    solution = name_exact_solution()
    fail = name_test_fail_variable()

    return ["",
            "// Maximum norm of difference between dof vectors of the",
            "// numerical solution and the interpolation of the exact solution",
            "using Dune::PDELab::Backend::native;",
            "double maxerror(0.0);",
            "for (std::size_t i=0; i<native({}).size(); ++i)".format(v),
            "  if (std::abs(native({})[i]-native({})[i]) > maxerror)".format(v, solution),
            "    maxerror = std::abs(native({})[i]-native({})[i]);".format(v, solution),
            "std::cout << std::endl << \"Maxerror: \" << maxerror << std::endl;",
            "if (maxerror>{})".format(get_option("compare_dofs")),
            "  {} = true;".format(fail)]


def type_discrete_grid_function(gfs):
    return "DGF_{}".format(gfs.upper())


@preamble
def define_discrete_grid_function(gfs, vector_name, dgf_name):
    dgf_type = type_discrete_grid_function(gfs)
    return ["using {} = Dune::PDELab::DiscreteGridFunction<decltype({}),decltype({})>;".format(dgf_type, gfs, vector_name),
            "{} {}({},{});".format(dgf_type, dgf_name, gfs, vector_name)]


def name_discrete_grid_function(gfs, vector_name):
    dgf_name = "{}_dgf".format(gfs)
    define_discrete_grid_function(gfs, vector_name, dgf_name)
    return dgf_name


@preamble
def typedef_difference_squared_adapter(name):
    sol = name_exact_solution()
    vector = name_vector(get_formdata())
    dgf = name_discrete_grid_function(gfs, vector)

    return 'using {} = Dune::PDELab::DifferenceSquaredAdapter<decltype({}), decltype({})>;'.format(name, sol, dgf)


def type_difference_squared_adapter():
    name = 'DifferenceSquaredAdapter'
    typedef_difference_squared_adapter(name)
    return name


@preamble
def define_difference_squared_adapter(name):
    t = type_difference_squared_adapter()
    sol = name_exact_solution()
    vector = name_vector(get_formdata())
    dgf = name_discrete_grid_function(gfs, vector)

    return '{} {}({}, {});'.format(t, name, sol, dgf)


def name_difference_squared_adapter():
    name = 'differencesquared_adapter'
    define_difference_squared_adapter(name)
    return name


@preamble
def accumulate_L2_squared():
    dsa = name_difference_squared_adapter()
    accum_error = name_accumulated_L2_error()

    include_file("dune/pdelab/gridfunctionspace/gridfunctionadapter.hh", filetag="driver")
    include_file("dune/pdelab/common/functionutilities.hh", filetag="driver")

    return ["// L2 error squared of difference between numerical",
            "// solution and the interpolation of exact solution",
            "  Dune::PDELab::integrateGridFunction({}, {}, 10);".format(dsa, accum_error),
            ]


@preamble
def define_accumulated_L2_error(name):
    return "double {}(0.0);".format(name)


def name_accumulated_L2_error():
    name = 'l2error'
    define_accumulated_L2_error(name)
    return name


@preamble
def compare_L2_squared():
    accumulate_L2_squared()

    accum_error = name_accumulated_L2_error()
    fail = name_test_fail_variable()
    return ["std::cout << \"l2errorsquared: \" << {} << std::endl;".format(accum_error),
            "if (std::isnan({0}) or std::abs({0})>{1})".format(accum_error, get_option("compare_l2errorsquared")),
            "  {} = true;".format(fail)]


@preamble
def return_statement():
    from dune.perftool.pdelab.driver.error import name_test_fail_variable
    fail = name_test_fail_variable()
    return "return {};".format(fail)
