from dune.perftool.generation import (backend,
                                      domain,
                                      function_mangler,
                                      get_global_context_value,
                                      iname,
                                      instruction,
                                      kernel_cached,
                                      loopy_class_member,
                                      temporary_variable,
                                      )

from dune.perftool.sumfact.tabulation import (quadrature_points_per_direction,
                                              name_oned_quadrature_points,
                                              name_oned_quadrature_weights,
                                              )
from dune.perftool.pdelab.argument import name_accumulation_variable
from dune.perftool.pdelab.geometry import (local_dimension,
                                           )
from dune.perftool.options import get_option

from loopy import CallMangleInfo
from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType

from pymbolic.primitives import (Call,
                                 Product,
                                 Subscript,
                                 Variable,
                                 )

import pymbolic.primitives as prim
import loopy as lp
import numpy as np


def nest_quadrature_loops(kernel, inames):
    insns = []
    for insn in kernel.instructions:
        if "quad" in insn.tags:
            insns.append(insn.copy(within_inames=insn.within_inames.union(frozenset(inames)),
                                   tags=insn.tags - frozenset({"quad"})
                                   )
                         )
        else:
            insns.append(insn)

    return kernel.copy(instructions=insns)


class BaseWeight(FunctionIdentifier):
    def __init__(self, accumobj):
        self.accumobj = accumobj

    def __getinitargs__(self):
        return (self.accumobj,)

    @property
    def name(self):
        return '{}.weight'.format(self.accumobj)


@function_mangler
def base_weight_function_mangler(target, func, dtypes):
    if isinstance(func, BaseWeight):
        return CallMangleInfo(func.name, (NumpyType(np.float64),), ())


def pymbolic_base_weight():
    """ This is the base weight that should be multiplied to the quadrature
    weight. With the fast DG assembler this will handle the weighting of the
    time discretization scheme.
    TODO: Introduce backend switch that uses above BaseWeight function
    """
    return 1.0


@backend(interface="quad_inames", name="sumfact")
@kernel_cached
def quadrature_inames(element):
    if element is None:
        names = tuple("quad_{}".format(d) for d in range(local_dimension()))
    else:
        from ufl import FiniteElement
        assert isinstance(element, FiniteElement)
        from dune.perftool.pdelab.driver import FEM_name_mangling
        names = tuple("quad_{}_{}".format(FEM_name_mangling(element), d) for d in range(local_dimension()))
    domain(names, quadrature_points_per_direction())
    return names


@iname(kernel="operator")
def constructor_quad_iname(name, d, bound):
    name = "{}_{}".format(name, d)
    domain(name, quadrature_points_per_direction(), kernel="operator")
    return name


def constructor_quadrature_inames(dim, num1d):
    name = "quadiname_dim{}_num{}".format(dim, num1d)
    return tuple(constructor_quad_iname(name, d, quadrature_points_per_direction()) for d in range(local_dimension()))


def define_recursive_quadrature_weight(visitor, name, dir):
    info = visitor.current_info[1]
    if info is None:
        element = None
    else:
        element = info.element
    iname = quadrature_inames(element)[dir]
    temporary_variable(name, shape=(), shape_impl=())
    instruction(expression=Product((recursive_quadrature_weight(dir + 1),
                                    Subscript(Variable(name_oned_quadrature_weights()),
                                              (Variable(iname),)
                                              ))
                                   ),
                assignee=Variable(name),
                forced_iname_deps=frozenset(quadrature_inames()[dir:]),
                forced_iname_deps_is_final=True,
                tags=frozenset({"quad"}),
                )


def recursive_quadrature_weight(visitor, dir=0):
    if dir == local_dimension():
        return pymbolic_base_weight()
    else:
        name = 'weight_{}'.format(dir)
        define_recursive_quadrature_weight(visitor, name, dir)
        return Variable(name)


def quadrature_weight(visitor):
    # Return non-precomputed version
    if not get_option("precompute_quadrature_info"):
        return recursive_quadrature_weight(visitor)

    dim = local_dimension()
    num1d = quadrature_points_per_direction()
    name = "quad_weights_dim{}_num{}".format(dim, num1d)

    # Add a class member
    loopy_class_member(name,
                       dtype=np.float64,
                       shape=(num1d,) * dim,
                       classtag="operator",
                       dim_tags=",".join(["f"] * dim),
                       managed=True,
                       potentially_vectorized=True,
                       )

    # Precompute it in the constructor
    inames = constructor_quadrature_inames(dim, num1d)
    instruction(assignee=prim.Subscript(prim.Variable(name), tuple(prim.Variable(i) for i in inames)),
                expression=prim.Product(tuple(Subscript(Variable(name_oned_quadrature_weights()), (prim.Variable(i),)) for i in inames)),
                within_inames=frozenset(inames),
                kernel="operator",
                )

    info = visitor.current_info[1]
    if info is None:
        element = None
    else:
        element = info.element
    return prim.Subscript(lp.symbolic.TaggedVariable(name, "operator_precomputed"), tuple(prim.Variable(i) for i in quadrature_inames(element)))


def define_quadrature_position(name, index):
    instruction(expression=Subscript(Variable(name_oned_quadrature_points()), (Variable(quadrature_inames()[index]),)),
                assignee=Subscript(Variable(name), (index,)),
                forced_iname_deps=frozenset(quadrature_inames(element)),
                forced_iname_deps_is_final=True,
                tags=frozenset({"quad"}),
                )


@backend(interface="quad_pos", name="sumfact")
def pymbolic_quadrature_position(index, visitor):
    # Return the non-precomputed version
    if not get_option("precompute_quadrature_info"):
        name = 'pos'
        temporary_variable(name, shape=(local_dimension(),), shape_impl=("fv",))
        define_quadrature_position(name, index)
        return prim.Subscript(prim.Variable(name), (index,))

    assert isinstance(index, int)
    dim = local_dimension()
    num1d = quadrature_points_per_direction()
    name = "quad_points_dim{}_num{}_dir{}".format(dim, num1d, index)

    loopy_class_member(name,
                       dtype=np.float64,
                       shape=(num1d,) * dim,
                       classtag="operator",
                       dim_tags=",".join(["f"] * dim),
                       managed=True,
                       potentially_vectorized=True,
                       )

    # Precompute it in the constructor
    inames = constructor_quadrature_inames(dim, num1d)
    instruction(assignee=prim.Subscript(prim.Variable(name), tuple(prim.Variable(i) for i in inames)),
                expression=Subscript(Variable(name_oned_quadrature_points()), (prim.Variable(inames[index]))),
                within_inames=frozenset(inames),
                kernel="operator",
                )

    info = visitor.current_info[1]
    if info is None:
        element = None
    else:
        element = info.element
    return prim.Subscript(lp.symbolic.TaggedVariable(name, "operator_precomputed"), tuple(prim.Variable(i) for i in quadrature_inames(element)))


@backend(interface="qp_in_cell", name="sumfact")
def pymbolic_quadrature_position_in_cell(restriction):
    from dune.perftool.pdelab.geometry import to_cell_coordinates
    return to_cell_coordinates(pymbolic_quadrature_position(), restriction)
